/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * digitalSTROM ds485-base library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-base library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-base library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <unistd.h>

#include "log.h"

/**
 * Glorious wrapper around POSIX getpeername
 * @socket
 * @in -- ipv4 address and port
 * @ret -- negative if error
 */
int get_peername(int socket, struct sockaddr_in *in)
{
    socklen_t alen = (socklen_t)sizeof(struct sockaddr_in);
    int ret;

    if (in == NULL) {
        return -EINVAL;
    }

    ret = getpeername(socket, (struct sockaddr *)in, &alen);
    if (ret < 0) {
        report(LOG_WARNING, "getpeername: %s(%d)\n", strerror(errno), errno);
        return -errno;
    }
    return 0;
}

int service_to_port(const char *service, const char *protocol)
{
    struct servent *pse;
    int port;

    if ((pse = getservbyname(service, protocol)) == NULL) {
        report(LOG_INFO, "Can't get \"%s\" service entry\n", service);
        return 0;
    }

    port = htons(ntohs((in_port_t)pse->s_port));
    return port;
}

int passivesock(int port, char *protocol, int qlen, int listenGlobal)
{
    struct protoent *ppe;
    struct sockaddr_in sin;
    int s, type, proto, one = 1;

    memset((char *) &sin, 0, sizeof(sin));
    sin.sin_family = (sa_family_t)AF_INET;

    if (listenGlobal)
        sin.sin_addr.s_addr = htonl(INADDR_ANY);
    else
        sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    sin.sin_port = htons(port);

    ppe = getprotobyname(protocol);
    if (strcmp(protocol, "udp") == 0) {
        type = SOCK_DGRAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_UDP;
    } else {
        type = SOCK_STREAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_TCP;
    }
    if ((s = socket(PF_INET, type, proto)) < 0) {
        int rval = errno;
        report(LOG_ERR, "Can't create socket %s\n", strerror(rval));
        return -rval;
    }
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,(int)sizeof(one)) == -1) {
        int rval = errno;
        report(LOG_ERR, "Error: SETSOCKOPT SO_REUSEADDR %s\n", strerror(rval));
        return -rval;
    }
    if (bind(s, (struct sockaddr *) &sin, (int)sizeof(sin)) < 0) {
        int rval = errno;
        report(LOG_ERR, "Can't bind to port %d %s\n", port, strerror(rval));
        if (rval == EADDRINUSE) {
            report(LOG_ERR, "Maybe daemon already running!\n");
        }
        return -rval;
    }
    if (type == SOCK_STREAM && listen(s, qlen) < 0) {
        int rval = errno;
        report(LOG_ERR, "Can't listen on port %d %s\n", port, strerror(rval));
        return -rval;
    }

    int opts = fcntl(s, F_GETFL);
    if (opts < 0) {
        int rval = errno;
        report(LOG_ERR, "fcntl: %s\n", strerror(rval));
        return -rval;
    }
    if (opts > 0) {
        if (fcntl(s, F_SETFL, opts | O_NONBLOCK) < 0) {
            int rval = errno;
            report(LOG_ERR, "fcntl: %s\n", strerror(rval));
            return -rval;
        }
    }

    one = IPTOS_LOWDELAY;
    if (setsockopt(s, IPPROTO_IP, IP_TOS, &one, sizeof one) < 0) {
        int rval = errno;
        report(LOG_ERR, "setsockopt: %s\n", strerror(rval));
        return -rval;
    }
    if (type == SOCK_STREAM) {
        one = 1;
        if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &one, sizeof one) < 0) {
            int rval = errno;
            report(LOG_ERR, "setsockopt: %s\n", strerror(rval));
            return -rval;
        }
    }

    return s;
}
