/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS485p, imported from libdSvDC.

    dS485p is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS485p is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include <avahi-client/client.h>
#include <avahi-client/lookup.h>

#include <avahi-common/error.h>
#include <avahi-common/domain.h>
#include <avahi-common/llist.h>
#include <avahi-common/malloc.h>
#include <avahi-common/simple-watch.h>
#include <avahi-common/timeval.h>

#include "discovery.h"
#include "base/log.h"
#include "base/util.h"
#include "base/ds_string.h"

static bool discover_ds485d;
static struct ds485_avahi_callbacks g_cb;

static int g_running;
static pthread_t avahi_tid;

typedef struct ds485p_avahi_handle {
    AvahiSimplePoll *simple_poll;
    AvahiClient *client;
} ds485p_avahi_handle_t;

typedef struct ServiceInfo ServiceInfo;
struct ServiceInfo {
    ds485p_avahi_handle_t *handle;
    AvahiIfIndex interface;
    AvahiProtocol protocol;
    char *name;
    char *type;
    char *domain;

    AvahiServiceResolver *resolver;

    AVAHI_LLIST_FIELDS(ServiceInfo, info);
};

static ServiceInfo *services = NULL;

static void create_resolver(ServiceInfo *info);

static ServiceInfo *add_service(ds485p_avahi_handle_t *handle,
                                AvahiIfIndex interface,
                                AvahiProtocol protocol,
                                const char *name,
                                const char *type,
                                const char *domain)
{
    struct ServiceInfo *i;

    i = avahi_new(struct ServiceInfo, 1);
    if (!i) {
        panic("OOM: service info new");
    }

    i->handle = handle;
    i->interface = interface;
    i->protocol = protocol;
    i->name = avahi_strdup(name);
    i->type = avahi_strdup(type);
    i->domain = avahi_strdup(domain);

    AVAHI_LLIST_PREPEND(ServiceInfo, info, services, i);
    return i;
}

static ServiceInfo *find_service(AvahiIfIndex interface,
                                AvahiProtocol protocol,
                                const char *name,
                                const char *type,
                                const char *domain)
{
    ServiceInfo *i;

    for (i = services; i; i = i->info_next) {
        if (i->interface == interface &&
            i->protocol == protocol &&
            strcasecmp(i->name, name) == 0 &&
            avahi_domain_equal(i->type, type) &&
            avahi_domain_equal(i->domain, domain)) {
            return i;
        }
    }

    return NULL;
}

static void remove_service(ServiceInfo *i) {
    assert(i);

    AVAHI_LLIST_REMOVE(ServiceInfo, info, services, i);

    if (i->resolver) {
        avahi_service_resolver_free(i->resolver);
    }

    avahi_free(i->name);
    avahi_free(i->type);
    avahi_free(i->domain);
    avahi_free(i);
}

static void ds485p_avahi_resolve_callback(
        AvahiServiceResolver *resolver,
        AVAHI_GCC_UNUSED AvahiIfIndex interface,
        AVAHI_GCC_UNUSED AvahiProtocol protocol,
        AvahiResolverEvent event,
        const char *name,
        const char *type,
        const char *domain,
        const char *host_name,
        const AvahiAddress *address,
        uint16_t port,
        AvahiStringList *txt,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    assert(resolver);
    assert(userdata);

    ServiceInfo *info = userdata;
    ds485p_avahi_handle_t *handle = info->handle;

    /* decouple info->resolver in case of retry */
    assert(info->resolver == resolver);
    info->resolver = NULL;

    /*
     * callback is called whenever a service has been
     * resolved successfully or timed out
     */
    switch (event) {
    case AVAHI_RESOLVER_FAILURE:
        if (avahi_client_errno(handle->client) == AVAHI_ERR_TIMEOUT) {
            /*
             * service consists of PTR/TXT/SRV DNS
             * descriptors, if one of them has shorter
             * time-to-live and expires earlier, resolving
             * fails. Keep retrying until success or
             * resolving fails with an error other than
             * timeout e.g. service has expired from cache
             * completely
             */
            report(LOG_NOTICE,
                   "resolver timeout:: service:\"%s\" type:\"%s\" domain:\"%s\"\n",
                   name, type, domain);
            create_resolver(info);
            break;
        }

        report(LOG_ERR,
               "resolver failure:: service:\"%s\" type:\"%s\" domain:\"%s\": err:%s\n",
               name, type, domain, avahi_strerror(avahi_client_errno(handle->client)));
        avahi_simple_poll_quit(handle->simple_poll);
        break;

    case AVAHI_RESOLVER_FOUND: {
        char ip[AVAHI_ADDRESS_STR_MAX];
        char *dsuid = NULL;
        avahi_address_snprint(ip, sizeof(ip), address);
        report(LOG_NOTICE, "found %s at: %s:%u (%s)\n", name, host_name,
               port, ip);

        AvahiStringList *pair = avahi_string_list_find(txt, "dSUID");
        if (pair) {
            dsuid = strchr((const char *)avahi_string_list_get_text(pair),
                          '=');
            if (dsuid) {
                // skip '='
                dsuid++;
            }
        }

        if (!dsuid) {
            report(LOG_DEBUG, "ignoring vdSM %s at: %s:%u (%s) - missing "
                   "dSUID in announcement\n", name, host_name, port, ip);
            break;
        }

        size_t buflen = /* tcp:// */ 6 + strlen(host_name) + /* : */ 1 +
                        /* 65535 */ 5 + /* NULL */ 1;
        char *uri = malloc(buflen);
        if (!uri) {
            report(LOG_ERR, "could not allocate memory for bus uri "
                   "(%s:%u / %s)\n", host_name, port, name);
            return;
        }
        snprintf(uri, buflen, "tcp://%s:%u", host_name, port);

        dsuid_t dsStruct;
        if (dsuid_from_string(dsuid, &dsStruct) != 0) {
            report(LOG_ERR, "failed to convert string to dsuid\n");
            return;
        }

        g_cb.bus_discovered(uri, name, &dsStruct);
        free(uri);
        break;
    }
    default:
        break;
    }

    avahi_service_resolver_free(resolver);
}

static void create_resolver(ServiceInfo *info)
{
    ds485p_avahi_handle_t *handle = info->handle;
    AvahiServiceResolver *resolver;
    /*
     * We ignore the returned resolver object. In the callback
     * function we free it. If the server is terminated before
     * the callback function is called the server will free
     * the resolver for us.
     */
    resolver = avahi_service_resolver_new(handle->client, info->interface,
                                          info->protocol, info->name, info->type,
                                          info->domain, AVAHI_PROTO_INET, 0,
                                          ds485p_avahi_resolve_callback, info);
    if (!resolver) {
        if (avahi_client_errno(handle->client) == AVAHI_ERR_NO_MEMORY) {
            panic("OOM: create resolver");
        }

        report(LOG_ERR, "create resolver failed service:\"%s\": err:%s\n ",
               info->name, avahi_strerror(avahi_client_errno(handle->client)));
        /* start over */
        avahi_simple_poll_quit(handle->simple_poll);
    }

    /*
     * could the callback have fired before, we get here?
     * we are running on the mainloop, so this is likely save
     */
    info->resolver = resolver;
}

static void ds485p_avahi_browse_callback(
        AvahiServiceBrowser *browser,
        AvahiIfIndex interface,
        AvahiProtocol protocol,
        AvahiBrowserEvent event,
        const char *name,
        const char *type,
        const char *domain,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    assert(browser);
    assert(userdata);

    ds485p_avahi_handle_t *avahi = userdata;

    /* called when a new services becomes available or is removed from LAN */

    switch (event) {
    case AVAHI_BROWSER_FAILURE:

        report(LOG_ERR, "avahi browse callback %s\n",
               avahi_strerror(avahi_client_errno(avahi_service_browser_get_client(browser))));
        avahi_simple_poll_quit(avahi->simple_poll);
        return;

    case AVAHI_BROWSER_NEW: {
        ServiceInfo *info;
        report(LOG_DEBUG,
               "browser: new service \"%s\" of type \"%s\" in domain \"%s\"\n",
               name, type, domain);

        if (find_service(interface, protocol, name, type, domain)) {
            return;
        }

        info = add_service(avahi, interface, protocol, name, type, domain);
        create_resolver(info);
        break;
    }

    case AVAHI_BROWSER_REMOVE: {
        ServiceInfo *info;

        report(LOG_DEBUG,
               "browser: remove service \"%s\" of type \"%s\" in domain '%s'\n",
               name, type, domain);

        info = find_service(interface, protocol, name, type, domain);
        if (!info) {
            return;
        }

        remove_service(info);
        g_cb.bus_left(name);
        break;
    }

    case AVAHI_BROWSER_ALL_FOR_NOW:
    case AVAHI_BROWSER_CACHE_EXHAUSTED:
        report(LOG_DEBUG, "browser: %s\n",
               event == AVAHI_BROWSER_CACHE_EXHAUSTED ? "CACHE_EXHAUSTED" : "ALL_FOR_NOW");
    default:
        break;
    }
}

static AvahiServiceBrowser *add_avahi_browser(ds485p_avahi_handle_t *handle,
                                              const char *stype)
{
    assert(handle->client);
    return avahi_service_browser_new(handle->client, AVAHI_IF_UNSPEC,
                                     AVAHI_PROTO_UNSPEC,
                                     stype,
                                     NULL,
                                     0,
                                     ds485p_avahi_browse_callback,
                                     handle);
}

static void ds485p_avahi_client_callback(
        AvahiClient *client,
        AvahiClientState state,
        void *d)
{
    assert(client);
    assert(d);

    /*
     * CAUTION: this callback might have fired before
     * userdata->client has been set
     */
    ds485p_avahi_handle_t *data = d;
    data->client = client; /* idempotent */

    switch (state) {
    case AVAHI_CLIENT_FAILURE:
        report(LOG_ERR, "avahi client: conection failure: %s\n",
               avahi_strerror(avahi_client_errno(client)));
        avahi_simple_poll_quit(data->simple_poll);
        break;

    case AVAHI_CLIENT_S_COLLISION:
        /* not reached: we do not advertize */
        report(LOG_ERR, "avahi-client: collision detected\n");
        break;

    case AVAHI_CLIENT_S_REGISTERING:
        /* daemon is starting up */
        report(LOG_DEBUG, "avahi-client: registering\n");
        break;

    case AVAHI_CLIENT_S_RUNNING:
        report(LOG_NOTICE, "avahi-client: running\n");

        if (!add_avahi_browser(data, "_ds-vdsm._tcp")) {
            report(LOG_ERR, "could not create avahi browser for vdc: %s\n",
                   avahi_strerror(avahi_client_errno(client)));
            avahi_simple_poll_quit(data->simple_poll);
        }

        if (discover_ds485d) {
            if (!add_avahi_browser(data, "_ds-485d._tcp")) {
                report(LOG_ERR, "could not create avahi browser for ds485d: %s\n",
                       avahi_strerror(avahi_client_errno(client)));
                avahi_simple_poll_quit(data->simple_poll);
            }
        }

        break;

    case AVAHI_CLIENT_CONNECTING:
        /*
         * We're still connecting. This state is only
         * entered when AVAHI_CLIENT_NO_FAIL has been passed
         * to avahi_client_new() and the daemon is not yet
         * available.
         */
        report(LOG_NOTICE, "avahi-client: connecting / server not available\n");
        break;
    }
}

static AvahiClient *create_client(ds485p_avahi_handle_t *data)
{
    AvahiClient *client;
    int error;

    /* client callback might fire before function completes */
    client = avahi_client_new(avahi_simple_poll_get(data->simple_poll),
                            AVAHI_CLIENT_NO_FAIL,
                            ds485p_avahi_client_callback,
                            data, &error);
    if (!client) {
        if (error == AVAHI_ERR_NO_MEMORY) {
            panic("OOM: avahi client create");
        } else if (error == AVAHI_ERR_DBUS_ERROR) {
            report(LOG_ERR, "avahi create client:: dbus err:%s\n",
                   avahi_strerror(error));
            return NULL;
        }
        /* shouldn't happen due to AVAHI_CLIENT_NO_FAIL */
        report(LOG_ERR, "avahi create client:: err: %s\n",
               avahi_strerror(error));
        return NULL;
    }

    return client;
}

static bool ds485p_avahi_init(ds485p_avahi_handle_t *avahi)
{
    avahi->simple_poll = NULL;
    avahi->client = NULL;

    avahi->simple_poll = avahi_simple_poll_new();
    if (!avahi->simple_poll) {
        report(LOG_ERR, "could not allocate avahi_poll object\n");
        return false;
    }

    avahi->client = create_client(avahi);
    if (!avahi->client) {
        /* already logged */
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
        return false;
    }

    return true;
}

void ds485p_avahi_shutdown(ds485p_avahi_handle_t *avahi)
{
    while (services) {
        /* has pointer to resolvers, delete before client */
        remove_service(services);
    }

    if (avahi->client) {
        /* will free browsers and resolvers as well */
        avahi_client_free(avahi->client);
        avahi->client = NULL;
    }

    if (avahi->simple_poll) {
        avahi_simple_poll_free(avahi->simple_poll);
    }
}

static void* sysAvahiThread(void *arg)
{
    ds485p_avahi_handle_t *avahi_handle = arg;

    while (g_running) {
        if (ds485p_avahi_init(avahi_handle)) {
            avahi_simple_poll_loop(avahi_handle->simple_poll);
            ds485p_avahi_shutdown(avahi_handle);
        }

        if (!g_running) {
          break;
        }

        report(LOG_WARNING,
               "avahi daemon interrupted or stopped... reconnect.\n");

        /*
         * If ds485p_avahi_init fails we will retry again and again.
         * To avoid a busy loop, we will wait a second.
         */
        sleep(1);
    }

    report(LOG_ERR, "exit avahi discovery\n");
    return NULL;
}

/* don't use these global other than start/stop discover */
static ds485p_avahi_handle_t g_XXX_avahi_handle;

int start_discovery(int filter, struct ds485_avahi_callbacks *cb)
{
    int ret;
    g_running = true;

    discover_ds485d = (filter & DISCOVER_DS485D);

    if (cb == NULL) {
        return -EINVAL;
    }
    memcpy(&g_cb, cb, sizeof(*cb));

    ret = pthread_create(&avahi_tid, NULL, sysAvahiThread, &g_XXX_avahi_handle);
    if (ret != 0) {
        report(LOG_ERR, "Could not create avahi thread.\n");
        return ret < 0 ? ret : -ret; /* make sure it's negative */
    }
    pthread_detach(avahi_tid);
    return 0;
}

void stop_discovery()
{
    g_running = false;
    avahi_simple_poll_quit(g_XXX_avahi_handle.simple_poll);
}
#endif//HAVE_AVAHI
