/*
 *
 This file is part of the digitalSTROM ds485-proxy-daemon (ds485p).

 Copyright (C) 2012 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-proxy-daemon is free software;
 you can redistribute it and/or modify it under the terms of the
 GNU General Public License as published by the Free Software
 Foundation version 2 of the License.

 digitalSTROM ds485-proxy-daemon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-daemon; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>
#include <utlist.h>
#include <uthash.h>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#ifdef HAVE_LIBCONFIG
#include <libconfig.h>
#include <sys/stat.h>
#endif

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/ds485-client-interface.h>
#include <digitalSTROM/ds485-socket-server.h>

#include "base/ds_string.h"
#include "base/log.h"
#include "base/socket_server.h"
#include "base/util.h"

#ifdef HAVE_AVAHI
#include "discovery.h"
#endif

/******************************************************************************
 *
 */

extern char *optarg;
extern int optind;

int mainActive = 1;
int gotSignal = 0;

/* process id file */
static char* pidfile = "/var/run/ds485p.pid";

/* listen on all network listen, or just loopback */
static int listenGlobal = 0;

/* listening service for client connections */
static const char* ds485_service = "ds485";
static int default_port = 8442;

/* default dsuid */
static dsuid_t dsuid_default = { {0x35, 0x04, 0x17, 0x5f, 0xe0, 0, 0, 0, 0, 0, 0, 0xf0, 0, 0, 0, 1, 0} };

/**
 * Bus Session
 */
typedef struct _dS485pBus {
    struct _dS485pBus* next;

    pthread_mutex_t connMutex; /* ds485-client library not threadsafe */
    ds485ClientHandle_t connHandle;
    char* connUri;
    char* connName;
    dsuid_t dsuid;

    bool hasLink;

#define DELAY_BUS_CHANGE_TILL_RESCAN        10
#define BUS_RESCAN_PERIOD                   180
    struct timeval last_change;
    struct timeval last_scan;

    pthread_mutex_t listMutex; /* list read/update from different threads */
    dsuid_t* devList;
    int devListSize;

    bool autodiscovered;
} dS485pBus;

static dS485pBus* busSessions = NULL;
static pthread_rwlock_t g_busSessionsRWLock = PTHREAD_RWLOCK_INITIALIZER;

typedef struct vdsm_whitelist_hash {
    dsuid_t dsuid;
    UT_hash_handle hh;
} vdsm_whitelist_hash_t;

static vdsm_whitelist_hash_t *g_autoconnect_whitelist = NULL;
bool has_whitelist;

/**
 * dS485 Device Data - map device to bus
 */
typedef struct _dS485pDevice {
    dsuid_t devId;
    dS485pBus* devBus;
    UT_hash_handle hh;
} dS485pDevice;

/**
 * Private Data
 */
typedef struct dS485p {

    pthread_t connThread; /* bus connection supervisor */
    pthread_t devicesThread; /* device list fetcher */

    /* ds485p has no dsuid, reuse the one form the first bus */
    dsuid_t dsuid;
    int bus_change_fifo[2];
    pthread_mutex_t fifoWriteMutex;

    dS485pDevice* devHash;
    struct dS485_server *netlibHandle;
} dS485pData;

static dS485pData pData;
static pthread_rwlock_t deviceListRWLock;

/*
 * LOCKING:
 *
 * always lock in this order
 * 1. g_busSessionsMutex
 * 2. deviceListRWLock
 * 3. dS485pBus.connMutex
 *
 * you don't have to start at 1) but never take a lock
 * with a lower number after taking one with a higher one
 *
 * NOTE:
 * we could create deadlocks if calls into netlib or ds485-client
 * trigger callbacks, e.g. client_send could trigger
 * link_state left when it detect the connection is broken.
 * -- currently this is safe since link_state/bus_change are
 * handled asynchronously without taking locks
 *
 * TODO maybe merge deviceListMutex / g_busSessionsMutex
 *
 */

/* Name of (unprivileged) user to change to when we drop privileges. */
#ifndef DS485P_USER
#define DS485P_USER    "nobody"
#endif


static int ds485p_message_callback(void *handle, const ds485_container_t *container);
static int ds485p_netlib_packet_callback(void *handle, const ds485n_packet_t *packet);
static int ds485p_link_callback(void *_handle, bool _link);
static int ds485p_bus_change_callback(void *handle, dsuid_t *id, int flags);
#ifdef HAVE_AVAHI
static void emit_avahi_bus_left(dS485pBus *bus);
#endif
/********************* SIGNAL & ERROR HANDLING *********************/

/*
 * Get the signal and save for main-loop processing
 */
void
onsignal(int signo)
{
    gotSignal = signo;
    mainActive = 0;

    /* shutdown socket interface */
    ds485_server_abort(pData.netlibHandle);
}


/********************* DS485 SERVER IMPL ****************************/
dsuid_t ds485_server_get_dsuid(struct dS485_server *server)
{
    (void)server;
    return pData.dsuid;
}

typedef struct {
    union {
        const ds485_container_t *container;
        const ds485n_packet_t *packet;
    } u;
    enum {
        v_container,
        v_packet
    } type;
} packet_union_t;

/**
 * CAUTION: g_busSessionsRWLock must be held
 */
static int _send_packet(dS485pBus *bus, const dsuid_t *dest,
                       packet_union_t *data)
{
    int rval;

    if (!bus->hasLink) {
        report(LOG_NOTICE, "send_packet %s: inactive bus: %s\n",
               dsuid_is_broadcast(dest) ? "broadcast" : "unicast",
               bus->connUri);
        return -EHOSTDOWN;
    }

    pthread_mutex_lock(&bus->connMutex);

    if (!bus->connHandle) {
        pthread_mutex_unlock(&bus->connMutex);
        report(LOG_INFO, "send_packet %s: unconnected bus:%s\n",
               dsuid_is_broadcast(dest) ? "broadcast" : "unicast",
               bus->connUri);
        return -EBADF;
    }

    rval = (data->type == v_container) ?
        ds485_client_send_command(bus->connHandle, data->u.container) :
        ds485_client_send_long_command(bus->connHandle, data->u.packet);

    pthread_mutex_unlock(&bus->connMutex);

    if (rval < 0) {
        report(LOG_NOTICE, "send_packet %s: bus:%s err:%s\n",
               dsuid_is_broadcast(dest) ? "broadcast" : "unicast",
               bus->connUri, ds485c_strerror(rval));
        return -ENETRESET;
    }
    return 0;
}

/**
 * bridge packet to other ds485 buses
 * if packet was received from ds485-client set src_bus to NULL
 * if packet war recevied from ds485d/vdSM or similar add that as src_bus
 *
 * @param dest -- destination address
 * @param data -- union of various formats
 * @param src_bus -- never forward data to this bus, can be NULL
 */
static int
ds485p_packet_bridge(const dsuid_t *dest, packet_union_t *data,
                     dS485pBus *src_bus)
{
    int rval;

    if (dsuid_is_broadcast(dest)) {
        /*
         * forward to all buses but the originating bus
         */
        dS485pBus* cb;
        pthread_rwlock_rdlock(&g_busSessionsRWLock);
        LL_FOREACH(busSessions, cb) {
            if (!src_bus || cb != src_bus) {
                (void)_send_packet(cb, dest, data); // best effort
            }
        }
        pthread_rwlock_unlock(&g_busSessionsRWLock);
    } else {
        /*
         * forward to bus closer to destination
         */
        dS485pDevice* cd;
        dS485pBus* dest_bus;

        pthread_rwlock_rdlock(&g_busSessionsRWLock);
        pthread_rwlock_rdlock(&deviceListRWLock);

        HASH_FIND(hh, pData.devHash, dest, sizeof(dsuid_t), cd);
        if (cd == NULL) {
            DSUID_STR(dsuid_str);
            dsuid_to_string(dest, dsuid_str);
            report(LOG_NOTICE, "unknown destination %s: src bus %s\n",
                   dsuid_str, src_bus ? src_bus->connUri: "client");

            pthread_rwlock_unlock(&deviceListRWLock);
            pthread_rwlock_unlock(&g_busSessionsRWLock);
            return -ENOENT;
        }
        dest_bus = cd->devBus;
        pthread_rwlock_unlock(&deviceListRWLock);

        if (src_bus && src_bus == dest_bus) {
            /*
             * message comes from the same bus that connects
             * to the destionation. And it's a request
             * otherwise we wouldn't bridge it. This
             * happens when that bus has another client.
             * This is a feature of the ds485-netlib to
             * forward all client requests to all other
             * clients.
             * It is a requirement of the dsampi. It's the
             * way its callback mechanism is implemented.
             * If a client submits a configuration change
             * request, all other clients get a heads up.
             * Think of cache-snooping, cache-coherence
             */
            DSUID_STR(dsuid_str);
            dsuid_to_string(dest, dsuid_str);
            report(LOG_NOTICE, "src bus: %s hosts destination %s (buddy client request)\n",
                   src_bus->connUri, dsuid_str);
            pthread_rwlock_unlock(&g_busSessionsRWLock); // src_bus
            return 0;
        }

        rval = _send_packet(dest_bus, dest, data);
        if (rval < 0) {
            pthread_rwlock_unlock(&g_busSessionsRWLock);
            report(LOG_NOTICE, "forwarding failed %s %s\n", dest_bus->connUri,
                   strerror(rval));
            return rval;
        }

        pthread_rwlock_unlock(&g_busSessionsRWLock);
    }

    /**
     * TODO, queue or drop
     * - what if broadcast to buses fail (call scene)?
     * - what if unicast packet fails?
     */
    return 0; /* always succeed */
}

/**
 * send to each downstream bus
 * */
int ds485_server_send_data(struct dS485_server *server,
                           const ds485_container_t *container)
{
    (void)server;
    packet_union_t data = {
        .u = { .container = container },
        .type = v_container
    };
    return ds485p_packet_bridge(&container->destinationId, &data, NULL);
}

int ds485_server_send_long_data(struct dS485_server *server,
                                const ds485n_packet_t* packet)
{
    (void)server;
    packet_union_t data = {
        .u = { .packet = packet },
        .type = v_packet
    };
    return ds485p_packet_bridge(&packet->destinationId, &data, NULL);
}

int ds485_server_get_bus_members(struct dS485_server *server,
                                 dsuid_t *devices, int length)
{
    int ct, n;
    (void)server;
    dS485pDevice *cur, *tmp;

    pthread_rwlock_rdlock(&deviceListRWLock);

    if ((unsigned)length < HASH_COUNT(pData.devHash)) {
        report(LOG_WARNING, "allocated device list too small, cropping %d < %d\n",
               length, HASH_COUNT(pData.devHash));
        n = length;
    } else {
        n = HASH_COUNT(pData.devHash);
    }

    ct = 0;
    HASH_ITER(hh, pData.devHash, cur, tmp) {
        if (ct >= n) {
            break;
        }
        memcpy(&devices[ct++], &cur->devId, sizeof(dsuid_t));
    }

    pthread_rwlock_unlock(&deviceListRWLock);
    return n;
}

int ds485_server_restart_bus(struct dS485_server *server)
{
    (void)server;
    /* ignore, nothing to do.*/
    return 0;
}

int ds485_server_set_packetfilter(struct dS485_server *server, uint8_t rxfilter)
{
    (void)server;
    (void)rxfilter;
    return 0; /* TODO */
}

struct dS485_server_callbacks netlib_cb =
{
    .receive_data = ds485_server_send_data,
    .receive_long_data = ds485_server_send_long_data,
    .get_bus_members = ds485_server_get_bus_members,
    .restart_bus = ds485_server_restart_bus,
    .get_dsuid = ds485_server_get_dsuid,
    .set_packetfilter = ds485_server_set_packetfilter,
};

/********************* DS485P HELPERS ****************************/

static dS485pBus *ds485p_create_bus(const char *url, const char *name,
                                    const dsuid_t *dsuid, bool autodiscovered)
{
    dS485pBus* cb = NULL;

    cb  = (dS485pBus*)calloc(1, sizeof(dS485pBus));
    if (!cb) {
        report(LOG_ERR, "create bus: out of memory %s \n", url);
        return NULL;
    }

    cb->connUri = strdup(url);
    if (!cb->connUri) {
        report(LOG_ERR, "create bus: out of memory %s \n", url);
        free(cb);
        return NULL;
    }

    cb->connName = strdup(name);
    if (!cb->connName) {
        report(LOG_ERR, "create bus: out of memory %s \n", url);
        free(cb->connUri);
        free(cb);
        return NULL;
    }

    if (dsuid) {
        memcpy(&cb->dsuid, dsuid, sizeof(dsuid_t));
    }

    cb->autodiscovered = autodiscovered;
    cb->hasLink = false;

    pthread_mutex_init(&cb->connMutex, NULL);
    pthread_mutex_init(&cb->listMutex, NULL);

    return cb;
}

static void ds485p_free_bus(dS485pBus *cb)
{
    if (!cb) {
        return;
    }

    pthread_mutex_lock(&cb->connMutex);
    ds485_client_close(cb->connHandle);
    pthread_mutex_unlock(&cb->connMutex);
    if (cb->connUri) {
        free(cb->connUri);
    }

    if (cb->connName) {
        free(cb->connName);
    }

    pthread_mutex_destroy(&cb->connMutex);
    pthread_mutex_destroy(&cb->listMutex);
    free(cb);
}

/**
 * Check if a given pointer to bus is still valid
 *
 * CAUTION: g_busSessionsRWLock must be held
 * TODO: malloc might return same before in subsequent join/leave cycle
 */
static dS485pBus *_ds485p_check_bus(dS485pBus *token)
{
    dS485pBus* cb;

    if (token == NULL) {
        return NULL;
    }
    LL_FOREACH(busSessions, cb) {
        if (cb == token) {
            return cb;
        }
    }
    return NULL;
}

/**
 * CAUTION: g_busSessionsRWLock must be held
 */
static dS485pBus *_ds485p_lookup_bus(const dsuid_t *dsuid)
{
    dS485pBus* cb;

    if (dsuid == NULL) {
        return NULL;
    }

    LL_FOREACH(busSessions, cb) {
        if (dsuid_equal(&cb->dsuid, dsuid)) {
            return cb;
        }
    }
    return NULL;
}

static int ds485p_add_bus(const char *uri, const char *name,
                           const dsuid_t *dsuid, bool autodiscovered)
{
    dS485pBus* newBus;

    if (!dsuid && autodiscovered) {
        report(LOG_ERR, "dsuid not valid \n");
        return -EINVAL;
    }

    newBus = ds485p_create_bus(uri, name, dsuid, autodiscovered);
    if (newBus == NULL) {
        report(LOG_NOTICE, "failed to create new bus\n");
        return -ENOMEM;
    }

    pthread_rwlock_wrlock(&g_busSessionsRWLock);

    if (dsuid && (_ds485p_lookup_bus(dsuid) != NULL)) {
        /*
         * bus may have been already added via a persistent connection
         * BOGUS: persistent connections have no dsuid set
         * TODO fetch dsuid of perstistent buses as well *
         */
        report(LOG_DEBUG, "bus %s already known\n", uri);
        pthread_rwlock_unlock(&g_busSessionsRWLock);
        ds485p_free_bus(newBus);
        return 0;
    }

    LL_APPEND(busSessions, newBus);

    pthread_rwlock_unlock(&g_busSessionsRWLock);
    return 0;
}

/** takes rwlock of g_busSessionsRWLock */
static int ds485p_remove_bus(dS485pBus *bus)
{
    pthread_rwlock_wrlock(&g_busSessionsRWLock);
    dS485pBus *cb, *tmp;

    LL_FOREACH_SAFE(busSessions, cb, tmp) {
        if (cb == bus) {
            report(LOG_DEBUG, "removing bus uri:%s avahi:%d\n",
                   cb->connUri, cb->autodiscovered);
            LL_DELETE(busSessions, cb);
            ds485p_free_bus(cb);
        }
    }
    pthread_rwlock_unlock(&g_busSessionsRWLock);

    return 0;
}

#ifdef HAVE_AVAHI
static void init_whitelist(const config_setting_t *vdsms)
{
    int i, ct = 0;
    if (!config_setting_is_array(vdsms)) {
        report(LOG_WARNING, "autoconnect-whitelist is not array\n");
        return;
    }

    for (i = 0; i < config_setting_length(vdsms); i++) {
        vdsm_whitelist_hash_t *item;
        // prevent double entries

        const char *dsuid_str = config_setting_get_string_elem(vdsms, i);
        if (!dsuid_str) {
            continue;
        }

        dsuid_t dsuid;
        if (dsuid_from_string(dsuid_str, &dsuid) != 0) {
            report(LOG_WARNING, "invalid dsuid %s\n", dsuid_str);
            continue;
        }

        HASH_FIND(hh, g_autoconnect_whitelist, &dsuid, sizeof(dsuid), item);
        if (item != NULL) {
            // already present
            continue;
        }

        item = malloc(sizeof(*item));
        if (!item) {
            report(LOG_ERR, "ENOMEM autoconnect whitelist\n");
            continue;
        }
        memcpy(&item->dsuid, &dsuid, sizeof(dsuid_t));
        HASH_ADD(hh, g_autoconnect_whitelist, dsuid, sizeof(dsuid), item);
        report(LOG_NOTICE, "whitelist: %s\n", dsuid_str);
        ct++;
    }
    report(LOG_NOTICE, "whitelist size: %d\n", ct);
}

static void avahi_bus_discovered(const char *uri, const char *name,
                                 const dsuid_t *dsuid)
{
    if (has_whitelist) {
        vdsm_whitelist_hash_t *tmp;

        HASH_FIND(hh, g_autoconnect_whitelist, dsuid, sizeof(dsuid_t), tmp);
        if (!tmp) {
            DSUID_STR(dsuid_str);
            dsuid_to_string(dsuid, dsuid_str);
            report(LOG_INFO, "bus %s not in the whitelist.... ingore\n", dsuid_str);
            return;
        }
    }

    ds485p_add_bus(uri, name, dsuid, true);
}

/**
 * TODO make avahi do the name->dsuid lookup
 */
static void avahi_bus_left(const char* name)
{
    dS485pBus *cb, *found = NULL;

    pthread_rwlock_rdlock(&g_busSessionsRWLock);
    LL_FOREACH(busSessions, cb) {
        if (cb->autodiscovered && cb->connName &&
            (strcmp(cb->connName, name) == 0)) {
            report(LOG_NOTICE, "avahi: %s left\n", cb->connUri);
            found = cb;
            break;
        }
    }
    pthread_rwlock_unlock(&g_busSessionsRWLock);

    if (!found) {
        report(LOG_NOTICE, "avahi bus left: bus not found %s\n", name);
        return;
    }
    emit_avahi_bus_left(found); // pointer will be checked again
}
#endif

static dS485pDevice *create_device_entry(dsuid_t dsuid, dS485pBus* bus)
{
    dS485pDevice *tmp = malloc(sizeof(dS485pDevice));
    if (!tmp) {
        report(LOG_WARNING, "%d malloc failed\n", __LINE__);
        return NULL;
    }
    tmp->devId = dsuid;
    tmp->devBus = bus;
    return tmp;
}

/**
 * CAUTION: read lock g_busSessionsRWLock must be held
 * CAUTION: listMutex lock must be held
 */
static dsuid_t *__bus_find_dsm(dS485pBus *bus, dsuid_t *dsuid)
{
    int i;
    for (i = 0; i < bus->devListSize; i++) {
        if (!memcmp(&bus->devList[i], dsuid, sizeof(*dsuid))) {
            return &bus->devList[i];
        }
    }
    return 0;
}

/**
 * CAUTION: read lock g_busSessionsRWLock must be held
 */
static int _bus_append_dsm(dS485pBus *bus, dsuid_t *dsuid)
{
    dsuid_t *tmp;
    DSUID_STR(dsuid_str);

    pthread_mutex_lock(&bus->listMutex);
    if (__bus_find_dsm(bus, dsuid)) {
        pthread_mutex_unlock(&bus->listMutex);
        dsuid_to_string(dsuid, dsuid_str);
        report(LOG_NOTICE, "%s dsuid already present %s\n", __func__, dsuid_str);
        return 0; /* done */
    }

    tmp = realloc(bus->devList, (bus->devListSize + 1) * sizeof(dsuid_t));
    if (!tmp) {
        /* original list is untouched though */
        dsuid_to_string(dsuid, dsuid_str);
        report(LOG_ERR, "%s realloc failed, not adding dsuid %s\n", __func__,
               dsuid_str);
        pthread_mutex_unlock(&bus->listMutex);
        return ENOMEM;
    }

    bus->devList = tmp;
    memcpy(&bus->devList[bus->devListSize], dsuid, sizeof(*dsuid));
    bus->devListSize++;
    pthread_mutex_unlock(&bus->listMutex);
    return 0;
}

/**
 * CAUTION: read lock g_busSessionsRWLock must be held
 */
static int _bus_remove_dsm(dS485pBus *bus, dsuid_t *dsuid)
{
    DSUID_STR(dsuid_str);
    int offset, rem;
    dsuid_t *victim;

    pthread_mutex_lock(&bus->listMutex);
    victim = __bus_find_dsm(bus, dsuid);
    if (!victim) {
        pthread_mutex_unlock(&bus->listMutex);
        dsuid_to_string(dsuid, dsuid_str);
        report(LOG_NOTICE, "%s dsuid not found %s\n", __func__, dsuid_str);
        return 0; /* done */
    }

    /*
     * | e0 | e1 | e2 | e3 | victim | e5 | e6 |  -> size == 7
     * | e0 | e1 | e2 | e3 | e5 | e6 |
     */
    offset = victim - bus->devList;
    rem = bus->devListSize - offset - 1;
    report(LOG_DEBUG, "%s offset:%d len: %d\n", __func__, offset,
           bus->devListSize - offset - 1);
    if (rem > 0) {
        memmove(victim, victim + 1, rem * sizeof(dsuid_t));
    }
    bus->devListSize--;
    pthread_mutex_unlock(&bus->listMutex);
    return 0;
}

static void print_usage(void)
{
    fprintf(stderr,
            "Usage: dS485p [OPTION] <uri> [<uri> ...]\n");
    fprintf(stderr,
            "       <uri> is a ds485d listening socket (example: tcp://localhost:1234)\n\n");
    fprintf(stderr,
            "       OPTION is a selection out of the following options:\n");
    fprintf(stderr,
            "       -h  --  print this help\n");
    fprintf(stderr,
            "       -d  --  debug-level: max debug level still logged: 0-7 (default: 5)\n");
    fprintf(stderr,
            "       -p  --  PID-file (default: /var/run/ds485p.pid)\n");
    fprintf(stderr,
            "       -F  --  stay in foreground (default: not set == daemonize)\n");
    fprintf(stderr,
            "               log-output: foreground: stdout\n");
    fprintf(stderr,
            "                           daemonized: syslog\n");
    fprintf(stderr,
            "       -G  --  enable external access: accept socket connection from other machines\n");
    fprintf(stderr,
            "       -l  --  listening port number for client connections (default: 8442)\n");
#ifdef HAVE_LIBCONFIG
    fprintf(stderr,
            "       -c  -- path to configuration file (default ./ds485p.cfg)\n");
#endif
    fprintf(stderr,
            "       -i  --  change dSUID of ds485p (default: 3504175fe0000000000000f00000000100\n");
    fprintf(stderr,
            "       -a  --  enable avahi discovery for ds485d\n");
    fprintf(stderr,
            "       -V  --  print ds485p version information\n");
    fprintf(stderr, "\n");
}


/************************ BUS SERVICE HANDLER ****************************/

struct bus_change {
    enum {
        DSM_CHANGED = 0,
        STATE_CHANGED,
        RESCAN_BUS,
        AVAHI_BUS_LEFT,
    } type;
    void *bus_token; /* use _ds485p_check_bus for validity before accessing */
    dsuid_t dsuid;

    union {
#define DSM_LEFT 0x1
        int flags;
        bool link;
    };
};

static void emit_link_change(dS485pBus *_bus, bool _link)
{
    struct bus_change msg;
    memset(&msg, 0, sizeof(msg));

    msg.type = STATE_CHANGED;
    msg.bus_token = _bus;
    msg.link = _link;

    pthread_mutex_lock(&pData.fifoWriteMutex);
    (void)write(pData.bus_change_fifo[1], &msg, sizeof(struct bus_change));
    pthread_mutex_unlock(&pData.fifoWriteMutex);
}

static void emit_rescan_bus(dS485pBus *bus)
{
    struct bus_change msg;
    memset(&msg, 0, sizeof(msg));

    msg.type = RESCAN_BUS;
    msg.bus_token = bus;

    pthread_mutex_lock(&pData.fifoWriteMutex);
    (void)write(pData.bus_change_fifo[1], &msg, sizeof(struct bus_change));
    pthread_mutex_unlock(&pData.fifoWriteMutex);
}

static void emit_dsm_changed(dS485pBus *bus, dsuid_t *dsuid, int flags)
{
    struct bus_change msg;
    memset(&msg, 0, sizeof(msg));

    msg.type = DSM_CHANGED;
    msg.bus_token = bus;
    memcpy(&msg.dsuid, dsuid, sizeof(dsuid_t));
    msg.flags = flags;

    pthread_mutex_lock(&pData.fifoWriteMutex);
    (void)write(pData.bus_change_fifo[1], &msg, sizeof(struct bus_change));
    pthread_mutex_unlock(&pData.fifoWriteMutex);
}

#ifdef HAVE_AVAHI
static void emit_avahi_bus_left(dS485pBus *bus)
{
    struct bus_change msg;
    memset(&msg, 0, sizeof(msg));

    msg.type = AVAHI_BUS_LEFT;
    msg.bus_token = bus;
    pthread_mutex_lock(&pData.fifoWriteMutex);
    (void)write(pData.bus_change_fifo[1], &msg, sizeof(struct bus_change));
    pthread_mutex_unlock(&pData.fifoWriteMutex);
}
#endif

/**
 * registered device callback for bus state changes
 */
static int
ds485p_link_callback(void *_handle, bool _link)
{
    dS485pBus* bus = (dS485pBus *)_handle;
    struct bus_change msg;
    memset(&msg, 0, sizeof(msg));

    report(LOG_NOTICE, "[link change] bus: %s: link: %i\n", bus->connUri,
           _link);
    emit_link_change(bus, _link);
    return 0;
}

static int
ds485p_bus_change_callback(void *handle, dsuid_t *id, int flags)
{
    dS485pBus* bus = (dS485pBus*) handle;
    DSUID_STR(dsuid_str);

    dsuid_to_string(id, dsuid_str);
    report(LOG_NOTICE, "[CB bus change] bus: %s dsuid: %s %s\n", bus->connUri,
           dsuid_str, (flags == DSM_LEFT) ? "left" : "joined");

    emit_dsm_changed(bus, id, flags);
    return 0;
}

/**
 * registered callback for frame reception
 */
static int
ds485p_message_callback(void *handle, const ds485_container_t *container)
{
    dS485pBus* bus = (dS485pBus*) handle;
    CONTAINER_STR_BUF(cont_buf);
    packet_union_t data = {
        .u = { .container = (ds485_container_t *)container },
        .type = v_container
    };

    sprintf_container(cont_buf, container);
    report(LOG_DEBUG, "Message: %s: %s\n", bus->connUri, cont_buf);
    if (container->length > DS485_MAX_PAYLOAD_SIZE) {
        report(LOG_WARNING, "Oversized message payload %d from bus %s\n",
               container->length, bus->connUri);
    }

    if (container->containerType == DS485_CONTAINER_REQUEST) {
        /*
         * ignore broadcasts from other devices and events/responses,
         * they have to stay locally on that bus
         */
        (void)ds485p_packet_bridge(&container->destinationId, &data, bus);
    }

    ds485_server_reply(pData.netlibHandle, container);
    return 0;
}

static int
ds485p_netlib_packet_callback(void *handle, const ds485n_packet_t *packet)
{
    dS485pBus* bus = (dS485pBus*) handle;
    CONTAINER_STR_BUF(cont_buf);
    packet_union_t data = {
        .u = { .packet = (ds485n_packet_t *)packet },
        .type = v_packet,
    };

    sprintf_netlib_packet(cont_buf, packet);
    report(LOG_DEBUG, "Long Message: %s: %s\n", bus->connUri, cont_buf);
    if (packet->length > DS485N_PACKET_MAX_PAYLOAD_SIZE) {
        report(LOG_WARNING, "Oversized long message payload %d from bus %s\n",
               packet->length, bus->connUri);
    }

    if (packet->containerType == DS485_CONTAINER_REQUEST) {
        /*
         * ignore broadcasts from other devices and events/responses,
         * they have to stay locally on that bus
         */
        (void)ds485p_packet_bridge(&packet->destinationId, &data, bus);
    }

    ds485_server_reply_long(pData.netlibHandle, packet);
    return 0;
}

/************************** BUS CONNECTION THREAD ***************************/

static void connect_buses()
{
    dS485pBus *cb;

    pthread_rwlock_rdlock(&g_busSessionsRWLock);
    LL_FOREACH(busSessions, cb) {
        ds485ClientHandle_t tempHandle;
        /*
         * Reconnects are handled by the ds485-client, here we deal
         * with those we could not connect at startup
         */
        pthread_mutex_lock(&cb->connMutex);
        if (cb->connHandle != NULL) {
            pthread_mutex_unlock(&cb->connMutex);
            continue;
        }
        pthread_mutex_unlock(&cb->connMutex);

        struct ds485c_callbacks callbacks = {
            .link_cb = ds485p_link_callback,
            .link_data = cb,
            .bus_change_cb = ds485p_bus_change_callback,
            .bus_change_data = cb,
            .container_pkt_cb = ds485p_message_callback,
            .container_pkt_data = cb,
            .netlib_pkt_cb = ds485p_netlib_packet_callback,
            .netlib_pkt_data = cb,
        };

        /*
         * do not hold the conn mutex during open, since that might
         * block other threads for a long time
         * TODO release busSessions lock so buses can be removed,
         * figure out how to ensure cb is still valid when reclaiming
         * the lock
         */
        tempHandle = ds485_client_open2(cb->connUri, 0, &callbacks);
        if (tempHandle == NULL) {
            report(LOG_ERR, "Cannot connect to ds485 bus at: %s\n",
                   cb->connUri);
            continue;
        }

        /* TODO fetch dsuid bus, see ds485p_add_bus */

        report(LOG_NOTICE, "Connected to bus %s\n", cb->connUri);

        pthread_mutex_lock(&cb->connMutex);
        if (cb->connHandle != NULL) {
            report(LOG_NOTICE, "Bus handle updated, close new handle %s\n",
                   cb->connUri);
            pthread_mutex_unlock(&cb->connMutex);
            ds485_client_close(tempHandle);
            continue;

        }
        cb->connHandle = tempHandle;
        pthread_mutex_unlock(&cb->connMutex);
    }
    pthread_rwlock_unlock(&g_busSessionsRWLock);
}

static void*
sysConnThread(void *arg __attribute__((unused)))
{
    dS485pBus *cb;
    int retval;

    while (mainActive) {

        sleep(30);

        struct timeval now;
        retval = gettimeofday(&now, 0);
        if (retval) {
            report(LOG_WARNING, "gettimeofday failed %s\n", strerror(errno));
        }

        /*
         * periodically rescan busses
         */
        pthread_rwlock_rdlock(&g_busSessionsRWLock);
        LL_FOREACH(busSessions, cb) {
            struct timeval diff;

            if (cb->connHandle == NULL || !cb->hasLink) {
                continue;
            }

            timersub(&now, &cb->last_change, &diff);
            if (diff.tv_sec < DELAY_BUS_CHANGE_TILL_RESCAN) {
                /*
                 * do not scan too often
                 * http://redmine.digitalstrom.org/issues/2140
                 */
                continue;
            }

            if (timercmp(&cb->last_change, &cb->last_scan, >)) {
                emit_rescan_bus(cb);
            } else {
                timersub(&now, &cb->last_scan, &diff);
                if (diff.tv_sec > BUS_RESCAN_PERIOD) {
                    emit_rescan_bus(cb);
                }
            }
        }
        pthread_rwlock_unlock(&g_busSessionsRWLock);

        /*
         * periodically rescan bus states
         */
        pthread_rwlock_rdlock(&g_busSessionsRWLock);
        LL_FOREACH(busSessions, cb) {
            bool link;
            int ret;

            if (cb->connHandle == NULL) {
                continue;
            }

            /*
             * ds485-client emits initial state immediately
             * upon connection, before callbacks are registered
             * hence we might loose the initial 'connected'
             * callback
             */
            {
                pthread_mutex_lock(&cb->connMutex);
                ret = ds485_client_query_link(cb->connHandle, &link);
                pthread_mutex_unlock(&cb->connMutex);
                // assume disconnected, if call failed
                link = (ret) ? false : link;
            }

            if (link == cb->hasLink) {
                // no change in link state
                continue;
            }

            report(LOG_NOTICE, "[sysConnThread]: detected state change %s %d\n",
                   cb->connUri, link);
            emit_link_change(cb, link);
        }
        pthread_rwlock_unlock(&g_busSessionsRWLock);

        /*
         * Reconnects are handled by the ds485-client, here we deal
         * with those we could not connect at startup
         */
        connect_buses();

    } /* mainActive */

    return NULL;
}

/**************************** DEVICE LIST THREAD ***************************/

static void update_global_device_list()
{
    dS485pBus* cb;
    dS485pDevice *fresh_hash;

    fresh_hash = NULL;

    pthread_rwlock_rdlock(&g_busSessionsRWLock);

    LL_FOREACH(busSessions, cb) {
        dS485pDevice *cd;
        int i;

        if ((cb->connHandle == NULL) || !cb->hasLink) {
            report(LOG_NOTICE, "Ignore devices on bus %s: state %d, handle: %p\n",
                   cb->connUri, cb->hasLink, cb->connHandle);
            continue;
        }

        pthread_mutex_lock(&cb->listMutex);
        report(LOG_NOTICE, "Advertize devices on bus %s: total %i\n",
               cb->connUri, cb->devListSize);

        for (i = 0; i < cb->devListSize; i++) {
            cd = create_device_entry(cb->devList[i], cb);
            if (!cd) {
                report(LOG_WARNING, "%d malloc failed\n", __LINE__);
                continue; /* TODO better idea? */
                /* return -ENOMEM; */
            }
            HASH_ADD(hh, fresh_hash, devId, sizeof(dsuid_t), cd);
        }
        pthread_mutex_unlock(&cb->listMutex);
    }
    pthread_rwlock_unlock(&g_busSessionsRWLock);

    {
        dS485pDevice *cur, *tmp, *old_hash, *cd;
        DSUID_STR(dsuid_str);

        pthread_rwlock_rdlock(&g_busSessionsRWLock);
        pthread_rwlock_wrlock(&deviceListRWLock);
        old_hash = pData.devHash;
        pData.devHash = fresh_hash;

        HASH_ITER(hh, old_hash, cur, tmp) {
            HASH_FIND(hh, fresh_hash, &cur->devId, sizeof(dsuid_t), cd);
            if (cd == NULL) {
                /* dsm left */
                dsuid_to_string(&cur->devId, dsuid_str);
                report(LOG_NOTICE, "emit %s dSUID %s left \n", cur->devBus->connUri, dsuid_str);
                ds485_server_emit_bus_change(pData.netlibHandle, cur->devId,
                                             DSM_LEFT);
            }
        }

        HASH_ITER(hh, fresh_hash, cur, tmp) {
            HASH_FIND(hh, old_hash, &cur->devId, sizeof(dsuid_t), cd);
            if (cd == NULL) {
                /* dsm joined */
                dsuid_to_string(&cur->devId, dsuid_str);
                report(LOG_NOTICE, "emit %s dSUID %s joined\n",
                       cur->devBus->connUri, dsuid_str);
                ds485_server_emit_bus_change(pData.netlibHandle, cur->devId, 0);
            }
        }
        pthread_rwlock_unlock(&deviceListRWLock);
        pthread_rwlock_unlock(&g_busSessionsRWLock);

        /* cleanup old_hash */
        HASH_ITER(hh, old_hash, cur, tmp) {
            cd = old_hash;
            HASH_DEL(old_hash, cd);
            free(cd);
        }
    }
}

/**
 * CAUTION: read lock g_busSessionsRWLock must be held
 */
static int _rescan_bus(dS485pBus* cb)
{
    dsuid_t dl[64];
    int num, n;

    if (!cb->hasLink) {
        /* would fail anyway */
        return -EHOSTDOWN;
    }

    if (cb->connHandle == NULL) {
        /* skip dead servers */
        return -EBADF;
    }

    report(LOG_NOTICE, "Query devices: %s\n", cb->connUri);

    pthread_mutex_lock(&cb->connMutex);
    num = ds485_client_query_devices(cb->connHandle, dl, 64);
    pthread_mutex_unlock(&cb->connMutex);

    pthread_mutex_lock(&cb->listMutex);
    if (num < 0) {
        report(LOG_WARNING, "Failed query devices of %s: err %s\n",
               cb->connUri, ds485c_strerror(num));
        pthread_mutex_unlock(&cb->listMutex);
        /* TODO: emit_link_change(cb, ) */
        return -ENETRESET;
    }

    /* realloc(p, size = 0) => free(p), return NULL */
    cb->devList = realloc(cb->devList, num * sizeof(dsuid_t));
    if (!cb->devList && num > 0) {
        /* original memory left untouched */
        int ret = -errno;
        report(LOG_ERR, "realloc failed, %s\n", strerror(-errno));
        pthread_mutex_unlock(&cb->listMutex);
        return ret;
    }

    cb->devListSize = num;
    for (n = 0; n < num; n++) {
        cb->devList[n] = dl[n];
    }

    (void)gettimeofday(&cb->last_scan, NULL);
    report(LOG_NOTICE, "Devices on bus %s: total %i\n", cb->connUri,
           cb->devListSize);

    pthread_mutex_unlock(&cb->listMutex);
    return 0;
}

static void*
sysDevicesThread(void *arg __attribute__((unused)))
{
    struct bus_change msg;
    int rd, retval;
    DSUID_STR(dsuid_str);
    fd_set rfds;

    while (mainActive) {
        FD_ZERO(&rfds);
        FD_SET(pData.bus_change_fifo[0], &rfds);

        /*
         * no periodic wake up required, unless we want to query
         * device list of busses periodically
         */
        retval = select(pData.bus_change_fifo[0] + 1, &rfds, NULL, NULL, NULL);

        if (retval == -1) {
            if (errno != EAGAIN)
                report(LOG_ERR, "sysDevicesThread -- select : %s\n", strerror(errno));
            continue;
        }

        assert(FD_ISSET(pData.bus_change_fifo[0], &rfds));

        rd = read(pData.bus_change_fifo[0], &msg, sizeof(struct bus_change));
        if (rd < 0 || rd != sizeof(struct bus_change)) {
            report(LOG_ERR, "sysDevicesThread read failed %s\n", strerror(errno));
            continue;
        }

        switch (msg.type) {
        case STATE_CHANGED: {
            dS485pBus *bus;

            pthread_rwlock_rdlock(&g_busSessionsRWLock);
            bus = _ds485p_check_bus(msg.bus_token);
            if (!bus) {
                report(LOG_DEBUG, "bus left, ignore state change\n");
                pthread_rwlock_unlock(&g_busSessionsRWLock);
                continue;
            }

            if (msg.link == bus->hasLink) {
                /* nothing to do */
                pthread_rwlock_unlock(&g_busSessionsRWLock);
                continue;
            }

            bus->hasLink = msg.link;
            report(LOG_NOTICE, "[sysDevicesThread state_change] %s: link %s\n",
                   bus->connUri, bus->hasLink ? "detected" : "lost");

            /*
             * dsm reboot must generate leave and join messages,
             * so clients know when to rescan dsm for changes
             */
            if (bus->hasLink) {
                _rescan_bus(bus);
            }

            pthread_rwlock_unlock(&g_busSessionsRWLock);
            update_global_device_list();
            break;
        }

        case DSM_CHANGED: {
            dS485pBus *bus;

            pthread_rwlock_rdlock(&g_busSessionsRWLock);
            bus = _ds485p_check_bus(msg.bus_token);
            if (!bus) {
                report(LOG_DEBUG, "bus left, ingore dsm change\n");
                pthread_rwlock_unlock(&g_busSessionsRWLock);
                continue;
            }

            dsuid_to_string(&msg.dsuid, dsuid_str);
            report(LOG_DEBUG, "[sysDevicesThread bus_change] Device %s %s\n",
                   dsuid_str, (msg.flags == DSM_LEFT) ? "left" : "joined");

            if (msg.flags & DSM_LEFT) {
                _bus_remove_dsm(bus, &msg.dsuid);
            } else {
                _bus_append_dsm(bus, &msg.dsuid);
            }

            /*
             * do not poll the bus immediately for full device list
             * see http://redmine.digitalstrom.org/issues/2140
             * leave it to the sysConnThread to do periodic scanning
             */
            (void)gettimeofday(&bus->last_change, NULL);
            pthread_rwlock_unlock(&g_busSessionsRWLock);

            /*
             * will forward device left message, always emit a message
             * even when the DSM only left for a little while, the
             * configuration might still have changed
             */
            update_global_device_list();
            break;
        }

        case RESCAN_BUS: {
            dS485pBus *bus;

            pthread_rwlock_rdlock(&g_busSessionsRWLock);
            bus = _ds485p_check_bus(msg.bus_token);
            if (!bus) {
                report(LOG_DEBUG, "bus left, ingore rescan bus\n");
                pthread_rwlock_unlock(&g_busSessionsRWLock);
                continue;
            }

            retval = _rescan_bus(bus);
            if (retval != 0) {
                report(LOG_NOTICE, "[sysDevicesThread] rescan failed %d, %s\n",
                       -retval, strerror(-retval));
            }
            pthread_rwlock_unlock(&g_busSessionsRWLock);

            update_global_device_list();
            break;
        }

        case AVAHI_BUS_LEFT: {
            dS485pBus *bus;

            /*
             * only mark the bus !active so that the "emit
             * dsm left" messages contain a valid bus
             * description, if we remove the bus before
             * emitting those messages the bus descriptor is
             * garbage
             */
            pthread_rwlock_rdlock(&g_busSessionsRWLock);
            bus = _ds485p_check_bus(msg.bus_token);
            if (!bus) {
                report(LOG_DEBUG, "bus left, ignore avahi bus left\n");
                pthread_rwlock_unlock(&g_busSessionsRWLock);
                continue;
            }
            bus->hasLink = false;
            pthread_rwlock_unlock(&g_busSessionsRWLock);

            update_global_device_list();
            ds485p_remove_bus(msg.bus_token);
            break;
        }}

    } /* mainActive */

    return NULL;
}

#if defined(HAVE_LIBCONFIG)
static void parse_config(const char *cfg_file)
{
    struct stat statbuf;
    config_t config;
    const config_setting_t *vdsms;

    if (!cfg_file) {
        return;
    }

    if (stat(cfg_file, &statbuf) != 0) {
        report(LOG_WARNING, "could not open configuration file %s\n", cfg_file);
        return;
    }

    config_init(&config);
    if (!config_read_file(&config, cfg_file)) {
        report(LOG_ERR, "%s: line %d %s\n", cfg_file,
               config_error_line(&config), config_error_text(&config));
        config_destroy(&config);
        return;
    }

#if defined(HAVE_AVAHI)
    vdsms = config_lookup(&config, "autoconnect-whitelist");
    if (vdsms) {
        has_whitelist = true;
        init_whitelist(vdsms);
    } else {
        report(LOG_INFO, "autoconnect-whitelist missing, connect to all buses\n");
    }
#endif

    config_destroy(&config);
}
#endif

/* close all bus connections and remove closed connection from session list */
static void close_all_bus_connections()
{
    /* close all bus connections */
    dS485pBus* cb;
    dS485pBus* tmp;

    pthread_rwlock_wrlock(&g_busSessionsRWLock);

    LL_FOREACH_SAFE(busSessions, cb, tmp) {
        LL_DELETE(busSessions, cb);
        ds485p_free_bus(cb);
    }

    pthread_rwlock_unlock(&g_busSessionsRWLock);
}

/******************************* MAIN *******************************/

/*
 * Function main (argc, )
 *
 *    Main function
 *
 */

int
main(int argc, char* argv[])
{
    int c, ret, port;
    int goBackground = 1;
    int debugLevel = 5;
    char *str_port = NULL;
    int dSUIdConfigured = 0;
    dsuid_t dsuid;
    int explicitClientConnection = 0;
    unsigned int discovery_filter = 0;
#ifdef HAVE_LIBCONFIG
    #define CFG_OPT "c:"
    const char *config_file = "ds485p.cfg";
#else
    #define CFG_OPT
#endif

    init_logger("ds485p", LOG_INFO, false);
    mainActive = 1;

    pthread_mutex_init(&pData.fifoWriteMutex, NULL);
    if (pipe(pData.bus_change_fifo)) {
        report(LOG_ERR, "failed to create pipe: %s\n", strerror(errno));
        exit(-1);
    }

    pthread_rwlock_init(&deviceListRWLock, NULL);

#ifdef HAVE_SETLOCALE
    setlocale(LC_NUMERIC, "C");
#endif

    if (argc < 2) {
        print_usage();
        exit(-1);
    }

    /* Look for options */
    while ((c = getopt(argc, argv, "had:p:FGl:i:V" CFG_OPT)) != -1) {
        switch (c) {
        case 'h':
            print_usage();
            exit(0);
        case 'd':
            debugLevel = atoi(optarg);
            break;
        case 'p':
            pidfile = strdup(optarg);
            break;
        case 'F':
            goBackground = 0;
            break;
        case 'G':
            listenGlobal = 1;
            break;
        case 'l':
            str_port = strdup(optarg);
            break;
#ifdef HAVE_LIBCONFIG
        case 'c':
            config_file = optarg;
            break;
#endif
        case 'i': {
            if (dsuid_from_string(optarg, &dsuid) != 0) {
                report(LOG_ERR, "dSUID in the wrong format\n");
                exit(-1);
            }
            dSUIdConfigured = 1;
            break;
        }
#ifdef HAVE_AVAHI
        case 'a': {
            discovery_filter |= DISCOVER_DS485D;
            break;
        }
#endif
        case 'V':
            printf("Version: v%s\n", VERSION);
            exit(0);
        default:
            exit(-1);
        }
    }

    set_log_level(debugLevel);

#if defined(HAVE_AVAHI) && defined(HAVE_LIBCONFIG)
    if (config_file) {
        parse_config(config_file);
    }
#endif

    /*------------------------------------------------------------------------*/
    /* assign client connections */
    for (; optind < argc; optind++) {
        /* posix function */
        if (access(argv[optind], R_OK)) {
            ++explicitClientConnection;
            ds485p_add_bus(argv[optind], argv[optind], NULL, false);
        }
    }

    /* go to background mode */
    if (goBackground) {
        enable_syslog(true);
        daemonize();
    }
    report(LOG_NOTICE, "%s, %s (built %s, %s)\n", PACKAGE_NAME, VERSION,
           __DATE__, __TIME__);

    /* PID file */
    if (pidfile) {
        FILE* fp;

        if ((fp = fopen(pidfile, "w")) != NULL) {
            fprintf(fp, "%u\n", (unsigned int) getpid());
            fclose(fp);
        } else {
            report(LOG_ERR, "Cannot create PID file: %s\n", pidfile);
        }
    }

    /* Trap signals */
    gotSignal = 0;
    signal(SIGINT, onsignal);
    signal(SIGTERM, onsignal);
    signal(SIGQUIT, onsignal);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);

    /*------------------------------------------------------------------------*/
    /* assign default dsuid of given dsuid (console) */
    if (dSUIdConfigured) {
        memcpy(&pData.dsuid, &dsuid, sizeof(dsuid_t));
    } else {
        memcpy(&pData.dsuid, &dsuid_default, sizeof(dsuid_t));
    }

    DSUID_STR(dsuid_str);
    dsuid_to_string(&pData.dsuid, dsuid_str);
    report(LOG_NOTICE, "dSUID is set to: %s\n", dsuid_str);

    /*------------------------------------------------------------------------*/
    /* register message callbacks to ensure a valid handle is set */
    pData.netlibHandle = ds485_server_init(&netlib_cb, true);
    if (!pData.netlibHandle) {
        report(LOG_ERR, "Failed to initialize netlib server\n");
        exit(-4);
    }

    /*------------------------------------------------------------------------*/
    /* avahi or at least one client has to be configured */
    if ((discovery_filter == 0) &&
        (explicitClientConnection == 0) ) {
        report(LOG_ERR, "Avahi or at least one client has to be configured\n");
        exit(-3);
    }

    /*------------------------------------------------------------------------*/
    /* Start connection thread */
    if (pthread_create(&pData.connThread, NULL, sysConnThread, NULL)) {
        report(LOG_ERR, "Could not create connection supervisor thread.\n");
        mainActive = 0;
        goto out;
    }
    pthread_detach(pData.connThread);

    /*------------------------------------------------------------------------*/
    /* Start devices thread */
    if (pthread_create(&pData.devicesThread, NULL, sysDevicesThread, NULL)) {
        report(LOG_ERR, "Could not create device list fetcher thread.\n");
        mainActive = 0;
        goto out;
    }
    pthread_detach(pData.devicesThread);


    /*------------------------------------------------------------------------*/
    /* Start avahi thread */
#ifdef HAVE_AVAHI
    struct ds485_avahi_callbacks cb = {
       .bus_discovered = avahi_bus_discovered,
       .bus_left = avahi_bus_left,
    };

    ret = start_discovery(discovery_filter, &cb);
    if (ret < 0) {
        report(LOG_ERR, "Could not create avahi thread\n");
        mainActive = 0;
        goto out;
    }
#endif

    if (str_port != NULL) {
        port = sscanf_port(str_port);
        if (port < 0) {
            report(LOG_ERR, "invalid port number %s, aborting...\n", str_port);
            exit(-2);
        }
    } else {
        port = service_to_port(ds485_service, "tcp");
        if (port == 0) {
            port = default_port;
        }
    }

    connect_buses();

    if ((ret = ds485_server_listen(pData.netlibHandle, port, listenGlobal))) {
        report(LOG_ERR, "Error listen for clients failed %d\n", ret);
        mainActive = 0;
        exit(2);
    }

out:
    switch (gotSignal) {
    case SIGTERM:
    case SIGINT:
        report(LOG_NOTICE, "got SIGTERM or SIGINT\n");
        mainActive = 0;
        stop_discovery();
        break;
    default:
        report(LOG_NOTICE, "got signal %d -- ignored\n", gotSignal);
        break;
    }

    ds485_server_abort(pData.netlibHandle);

    close_all_bus_connections();

    close(pData.bus_change_fifo[0]);
    close(pData.bus_change_fifo[1]);
    pthread_mutex_destroy(&pData.fifoWriteMutex);
    pthread_rwlock_destroy(&deviceListRWLock);
    pthread_rwlock_destroy(&g_busSessionsRWLock);

    ds485_server_cleanup(pData.netlibHandle);
    cleanup_logger();

    return 0;
}
